@@ -4,7 +4,7 @@ use strict;
 use warnings;
 use JSON::Validator::OpenAPI;
 
-our $VERSION = "0.04";
+our $VERSION = "0.05";
 use constant DEBUG => $ENV{SQLTP_OPENAPI_DEBUG};
 use String::CamelCase qw(camelize decamelize wordsplit);
 use Lingua::EN::Inflect::Number qw(to_PL to_S);
@@ -41,7 +41,10 @@ sub _debug {
 # heuristic 1: strip out single-item objects
 sub _strip_thin {
   my ($defs) = @_;
-  my @thin = grep { keys(%{ $defs->{$_}{properties} }) == 1 } keys %$defs;
+  my @thin = grep {
+    my @props = keys %{ $defs->{$_}{properties} };
+    @props == 1 or (@props == 2 and grep /count/i, @props)
+  } keys %$defs;
   if (DEBUG) {
     _debug("OpenAPI($_) thin, ignoring", $defs->{$_}{properties})
       for sort @thin;
@@ -175,26 +178,36 @@ sub _make_fk {
   );
 }
 
+sub _get_entity {
+  my ($schema, $name, $view2real) = @_;
+  $schema->get_table($name) || $schema->get_table($view2real->{$name});
+}
+
 sub _fk_hookup {
-  my ($schema, $fromtable, $fromkey, $totable, $tokey, $required) = @_;
+  my ($schema, $fromtable, $fromkey, $totable, $tokey, $required, $view2real) = @_;
   DEBUG and _debug("_fk_hookup $fromtable.$fromkey $totable.$tokey $required");
   my $from_obj = $schema->get_table($fromtable);
-  my $to_obj = $schema->get_table($totable);
+  my $to_obj = _get_entity($schema, $totable, $view2real);
   my $tokey_obj = $to_obj->get_field($tokey);
   my $field = $from_obj->get_field($fromkey) || $from_obj->add_field(
     name => $fromkey, data_type => $tokey_obj->data_type,
   );
   die $from_obj->error if !$field;
-  _make_fk($from_obj, $field, $totable, $tokey);
+  _make_fk($from_obj, $field, $to_obj->name, $tokey);
   _make_not_null($from_obj, $field) if $required;
   $field;
 }
 
 sub _def2table {
-  my ($name, $def, $schema, $m2m) = @_;
+  my ($name, $def, $schema, $m2m, $view2real) = @_;
   my $props = $def->{properties};
   my $tname = _def2tablename($name);
   DEBUG and _debug("_def2table($name)($tname)($m2m)", $props);
+  if (my $view_of = $def->{'x-view-of'}) {
+    my $target_table = _def2tablename($view_of);
+    $view2real->{$tname} = $target_table;
+    return (undef, []);
+  }
   my $table = $schema->add_table(
     name => $tname, comments => $def->{description},
   );
@@ -497,15 +510,21 @@ sub _make_many2many {
     for my $n2 (sort keys %{ $m2m{$n1} }) {
       my ($f1, $f2) = @{ $m2m{$n1}{$n2} };
       my ($t1_obj, $t2_obj) = map $schema->get_table($_->{to}), $f1, $f2;
+      my $f1_fromkey = $f1->{fromkey};
+      my $f2_fromkey = $f2->{fromkey};
+      if ($f1_fromkey eq $f2_fromkey and $f1->{from} eq $f2->{from}) {
+        $f1_fromkey =~ s#_id$#_to_id#;
+        $f2_fromkey =~ s#_id$#_from_id#;
+      }
       my ($table) = _def2table(
         $n1.$n2,
         {
           type => 'object',
           properties => {
-            $f1->{fromkey} => {
+            $f1_fromkey => {
               type => $SQL2TYPE{$t1_obj->get_field($f1->{tokey})->data_type}
             },
-            $f2->{fromkey} => {
+            $f2_fromkey => {
               type => $SQL2TYPE{$t2_obj->get_field($f2->{tokey})->data_type}
             },
           },
@@ -517,13 +536,13 @@ sub _make_many2many {
         to => $f1->{from},
         tokey => 'id',
         from => $table->name,
-        fromkey => $f1->{fromkey},
+        fromkey => $f1_fromkey,
         required => 1,
       }, {
         to => $f2->{from},
         tokey => 'id',
         from => $table->name,
-        fromkey => $f2->{fromkey},
+        fromkey => $f2_fromkey,
         required => 1,
       };
     }
@@ -538,6 +557,20 @@ sub _make_many2many {
   \@newfixups;
 }
 
+sub _remove_fields {
+  my ($defs, $name) = @_;
+  DEBUG and _debug("OpenAPI._remove_fields($name)", $defs);
+  for my $defname (sort keys %$defs) {
+    my $theseprops = $defs->{$defname}{properties} || {};
+    DEBUG and _debug("OpenAPI._remove_fields(t)($defname)", $theseprops);
+    for my $propname (keys %$theseprops) {
+      my $thisprop = $theseprops->{$propname};
+      DEBUG and _debug("OpenAPI._remove_fields(p)($propname)", $thisprop);
+      delete $theseprops->{$propname} if $thisprop->{$name};
+    }
+  }
+}
+
 sub parse {
   my ($tr, $data) = @_;
   my $openapi_schema = JSON::Validator::OpenAPI->new->schema($data)->schema;
@@ -548,6 +581,8 @@ sub parse {
   my @thin = _strip_thin(\%defs);
   DEBUG and _debug("thin ret", \@thin);
   delete @defs{@thin};
+  _remove_fields(\%defs, 'x-artifact');
+  _remove_fields(\%defs, 'x-input-only');
   %defs = %{ _merge_allOf(\%defs) };
   my $def2mask = defs2mask(\%defs);
   my $reffed = _find_referenced(\%defs);
@@ -557,16 +592,16 @@ sub parse {
   delete @defs{@subset};
   %defs = %{ _extract_objects(\%defs) };
   %defs = %{ _extract_array_simple(\%defs) };
-  my (@fixups);
+  my (@fixups, %view2real);
   %defs = %{ _fixup_addProps(\%defs) };
   %defs = %{ _absorb_nonobject(\%defs) };
   for my $name (sort keys %defs) {
-    my ($table, $thesefixups) = _def2table($name, $defs{$name}, $schema, 0);
+    my ($table, $thesefixups) = _def2table($name, $defs{$name}, $schema, 0, \%view2real);
     push @fixups, @$thesefixups;
   }
   my ($newfixups) = _make_many2many(\@fixups, $schema);
   for my $fixup (@$newfixups) {
-    _fk_hookup($schema, @{$fixup}{qw(from fromkey to tokey required)});
+    _fk_hookup($schema, @{$fixup}{qw(from fromkey to tokey required)}, \%view2real);
   }
   1;
 }
@@ -617,7 +652,8 @@ To try to make the data model represent the "real" data, it applies heuristics:
 =item *
 
 to remove object definitions that only have one property (which the
-author calls "thin objects")
+author calls "thin objects"), or that have two properties, one of whose
+names has the substring "count" (case-insensitive).
 
 =item *
 
@@ -704,6 +740,33 @@ in the definitions. Not exported. E.g.
     d2 => (1 << 1) | (1 << 2),
   }
 
+=head1 OPENAPI SPEC EXTENSIONS
+
+=head2 C<x-view-of>
+
+Under C</definitions/$defname>, a key of C<x-view-of> will name another
+definition (NB: not a full JSON pointer). That will make C<$defname>
+not be created as a table. The handling of creating the "view" of the
+relevant table is left to the CRUD implementation. This gives it scope
+to use things like the current requesting user, or web parameters,
+which otherwise would require a parameterised view. These are not widely
+available.
+
+=head2 C<x-artifact>
+
+Under C</definitions/$defname/properties/$propname>, a key of
+C<x-artifact> with a true value will indicate this is not to be stored,
+and will not cause a column to be created. The value will instead be
+derived by other means. The value of this key may become the definition
+of that derivation.
+
+=head2 C<x-input-only>
+
+Under C</definitions/$defname/properties/$propname>, a key of
+C<x-input-only> with a true value will indicate this is not to be stored,
+and will not cause a column to be created. This may end up being merged
+with C<x-artifact>.
+
 =head1 DEBUGGING
 
 To debug, set environment variable C<SQLTP_OPENAPI_DEBUG> to a true value.
